from axiom.store import Store
from functools import partial
from merlyn.multiplexing import addToStore, FactoryDict
from twisted.trial.unittest import SynchronousTestCase


class Factory(object):
    def __init__(self, store):
        self.store = store



prefix = "merlyn.test.test_multiplexing."
name = prefix + "Factory"



class OtherFactory(Factory):
    """Another factory.

    """



otherName = prefix + "OtherFactory"



class FactoryDictTests(SynchronousTestCase):
    """FactoryDicts are like dicts of names and identifiers except in an
    axiom store.

    """
    def setUp(self):
        self.store = Store()
        self.factoryDict = FactoryDict(self.store)
        addToStore(self.store, "test", name)


    def test_get(self):
        """FactoryDicts supply factories by identifiers. These factories are
        instantiated with the store.

        """
        self.checkFactory(self.factoryDict["test"])


    def checkFactory(self, factory, cls=Factory):
        """Checks the given factory. It must be an instance of the given class
        (default: Factory), as well as have the test store as the
        ``store`` attribute.

        """
        self.assertTrue(isinstance(factory, cls))
        self.assertIdentical(factory.store, self.store)


    def test_getMissing(self):
        """FactoryDicts raise KeyError when there is no factory with the given
        identifier.

        """
        self.assertRaises(KeyError, lambda: self.factoryDict["BOGUS"])


    def test_updateInStore(self):
        """The addToStore function will update a peristed factory's name, if a
        persisted factory with the passed identifier already exists.

        """
        addToStore(self.store, "test", otherName)
        self.checkFactory(self.factoryDict["test"], cls=OtherFactory)



import os

from axiom.store import Store
from clarent.exercise import GetExercises, GetExerciseDetails
from clarent.exercise import UnknownExercise, NotifySolved
from merlyn.exercise import Exercise, Locator, solveAndNotify
from merlyn.exercise import SolvableResourceMixin, Secret
from merlyn.auth import User
from twisted.trial.unittest import SynchronousTestCase
from txampext.respondertests import ResponderTestMixin



defaultExerciseKwargs = {
    "identifier": b"identifier",
    "title": u"\N{SNOWMAN}",
    "description": u"\N{CLOUD}"
}


def makeExercise(**kwargs):
    kw = defaultExerciseKwargs.copy()
    kw.update(kwargs)
    return Exercise(**kw)



class ExerciseTests(SynchronousTestCase):
    def test_wasSolvedBy(self):
        """The exercise knows when it was solved by a particular user.

        """
        store = Store()
        exercise = makeExercise(store=store)

        someUser = User(store=store, email="foo@example.com")
        self.assertFalse(exercise.wasSolvedBy(someUser))

        exercise.solvedBy(someUser)
        self.assertTrue(exercise.wasSolvedBy(someUser))

        someOtherUser = User(store=store, email="bar@example.com")
        self.assertFalse(exercise.wasSolvedBy(someOtherUser))



class SolveAndNotifyTests(SynchronousTestCase):
    def setUp(self):
        self.store = store = Store()
        self.user = User(store=store, email="test@example.com")
        self.proto = FakeProto(store, self.user)
        self.exercise = makeExercise(store=store)


    def wasSolved(self):
        """The current exercise has been marked as sovled by the current user.

        """
        return self.exercise.wasSolvedBy(self.user)


    def wasNotified(self):
        if self.proto.command is not NotifySolved:
            return False

        expectedKwargs = {
            b"identifier": b"identifier",
            b"title": u"\N{SNOWMAN}"
        }

        return self.proto.kwargs == expectedKwargs


    def test_solveAndNotify(self):
        self.assertFalse(self.wasSolved())
        self.assertFalse(self.wasNotified())

        solveAndNotify(self.proto, self.exercise)

        self.assertTrue(self.wasSolved())
        self.assertTrue(self.wasNotified())


    def test_resourceMixin(self):
        resource = SolvableResource(self.store)
        request = FakeRequest()
        request.transport = FakeTransport()
        request.transport.remote = self.proto

        self.assertFalse(self.wasSolved())
        self.assertFalse(self.wasNotified())

        resource.solveAndNotify(request)

        self.assertTrue(self.wasSolved())
        self.assertTrue(self.wasNotified())



class FakeProto(object):
    def __init__(self, store, user):
        self.store = store
        self.user = user

        self.command = self.kwargs = None


    def callRemote(self, command, **kwargs):
        self.command = command
        self.kwargs = kwargs



class FakeRequest(object):
    pass



class FakeTransport(object):
    pass



class SolvableResource(SolvableResourceMixin):
    exerciseIdentifier = b"identifier"



class _LocatorTests(object):
    def setUp(self):
        self.locator = Locator()
        self.locator.store = store = Store()
        self.locator.user = user = User(store=store, email=b"x@y.z")

        one = makeExercise(store=store, title=u"Exercise 1", identifier=b"1")
        one.solvedBy(user)

        makeExercise(store=store, title=u"Exercise 2", identifier=b"2")
        makeExercise(store=store, title=u"Exercise 3", identifier=b"3")



class GetExercisesTests(_LocatorTests, SynchronousTestCase):
    def test_getExercises(self):
        """A user can get some new exercises.

        """
        response = self.locator.getExercises(solved=False)
        exercises = list(response["exercises"])
        exercises.sort(key=lambda d: d[b"identifier"])
        self.assertEqual(exercises, [
            {b"title": u"Exercise 2", b"identifier": b"2"},
            {b"title": u"Exercise 3", b"identifier": b"3"}
        ])


    def test_getSolvedExercises(self):
        """A user can get previously solved exercises.

        """
        response = self.locator.getExercises(solved=True)
        exercises = list(response["exercises"])
        exercises.sort(key=lambda d: d["identifier"])
        self.assertEqual(exercises, [
            {b"title": u"Exercise 1", b"identifier": b"1"},
        ])



class GetExerciseDetailsTests(_LocatorTests, SynchronousTestCase):
    def test_getSolvedExerciseDetails(self):
        """A user can get details about a solved exercise.

        """
        details = self.locator.getExerciseDetails(identifier=b"1")
        self.assertEqual(details, {
            b"identifier": b"1",
            b"title": u"Exercise 1",
            b"description": u"\N{CLOUD}",
            b"solved": True
        })


    def test_getUnsolvedExerciseDetails(self):
        """A user can get details about an unsolved exercise.

        """
        details = self.locator.getExerciseDetails(identifier=b"2")
        self.assertEqual(details, {
            b"identifier": "2",
            b"title": u"Exercise 2",
            b"description": u"\N{CLOUD}",
            b"solved": False
        })


    def test_missingExercise(self):
        """When attempting to get details for an exercise that doesn't exist,
        an error is raised.

        """
        self.assertRaises(UnknownExercise,
                          self.locator.getExerciseDetails, identifier="BOGUS")



locator = Locator()



class GetExercisesResponderTests(ResponderTestMixin, SynchronousTestCase):
    command =  GetExercises
    locator = locator



class GetExerciseDetailsResponderTests(ResponderTestMixin, SynchronousTestCase):
    command =  GetExerciseDetails
    locator = locator



class SecretTests(SynchronousTestCase):
    def setUp(self):
        self.store = Store()
        self.user = User(store=self.store, email="test@example.com")


    def test_new(self):
        """When the user has no secret yet, gets a new secret.

        Asserts that the secret requests a sufficient number of bytes
        from urandom.

        """
        self.patch(os, "urandom", self._urandom)
        self.assertEqual(self.store.query(Secret).count(), 0)

        secret = Secret.forUser(self.user)
        self.assertEqual(secret.entropy, "sikrit")
        self.assertEqual(secret.user, self.user)


    def _urandom(self, nBytes):
        """
        Asserts that 256 bits worth of entropy is being requested.
        """
        self.assertEqual(nBytes, 256 // 8)
        return "sikrit"


    def test_same(self):
        """When the user already has a secret, gets that secret.

        """
        secret = Secret(store=self.store, entropy="xyzzy", user=self.user)
        self.assertIdentical(Secret.forUser(self.user), secret)


    def test_indexed(self):
        """The user attribute of secrets is indexed.

        """
        self.assertTrue(Secret.user.indexed)



from axiom.store import Store
from merlyn import auth, service
from twisted.conch.manhole_ssh import ConchFactory
from twisted.internet.protocol import connectionDone
from twisted.test.proto_helpers import MemoryReactor, StringTransport
from twisted.trial.unittest import SynchronousTestCase


class AMPTests(SynchronousTestCase):
    """Test cases for the AMP protocol class and its factory.

    """
    def setUp(self):
        self.store = Store()
        self.factory = service.Factory(self.store)


    def test_keepReference(self):
        """The protocol registers itself to its factory when a connection is
        made. The reference is removed when the connection is lost.

        """
        self.assertEqual(self.factory.protocols, set([]))

        proto = self.factory.buildProtocol(None)
        proto.makeConnection(StringTransport())
        proto.connectionMade()
        self.assertEqual(self.factory.protocols, set([proto]))

        proto.connectionLost(connectionDone)
        self.assertEqual(self.factory.protocols, set([]))


    def test_factoryHasStore(self):
        """The factory exposes its store as the ``store`` attribute.

        """
        self.assertIdentical(self.factory.store, self.store)


    def test_protocolsHaveStore(self):
        """Protocols have a reference to a store, which is the factory's store.

        """
        proto = self.factory.buildProtocol(None)
        self.assertIdentical(proto.store, self.store)



class ServiceTests(SynchronousTestCase):
    def setUp(self):
        self.store = Store()
        self.reactor = MemoryReactor()
        self.service = service.Service(self.store, reactor=self.reactor)


    def test_startService(self):
        """The service starts an AMP factory as well as a manhole.

        """
        self.service.startService()

        ampListenEvent, = self.reactor.sslServers
        port, factory, ctxFactory, _backlog, interface = ampListenEvent
        self.assertEqual(port, 4430)
        self.assertTrue(isinstance(factory, service.Factory))
        self.assertTrue(isinstance(ctxFactory, auth.ContextFactory))
        self.assertEqual(interface, "")

        manholeListenEvent, = self.reactor.tcpServers
        port, factory, _backlog, interface = manholeListenEvent
        self.assertEqual(port, 8888)
        self.assertTrue(isinstance(factory, ConchFactory))
        self.assertEqual(interface, "localhost")



class ServiceMakerTests(SynchronousTestCase):
    def test_options(self):
        """The service maker uses the merlyn Options class.

        """
        self.assertIdentical(service.ServiceMaker.options, service.Options)


    def test_tapname(self):
        """The service maker uses the tapname ``merlyn``.

        """
        self.assertEqual(service.ServiceMaker.tapname, b"merlyn")


    def test_makeService(self):
        """The service maker grabs the store from the options, and passes it
        to ``service.Service``.

        """
        maker = service.ServiceMaker()
        store = Store()
        svc = maker.makeService({"store": store})
        self.assertTrue(isinstance(svc, service.Service))
        self.assertIdentical(svc.store, store)



"""Exercises and helpers for exercise implementations.

"""
import os

from axiom import attributes, item, queryutil as q
from axiom.errors import ItemNotFound
from clarent import exercise as ce
from twisted.protocols import amp


class Exercise(item.Item):
    """An exercise.

    """
    identifier = attributes.bytes(allowNone=False)
    title = attributes.text(allowNone=False)
    description = attributes.text(allowNone=False)

    def solvedBy(self, user):
        """Stores that this user has just solved this exercise.

        You probably want to notify the user when this happens. For
        that, see ``solveAndNotify``.

        """
        _Solution(store=self.store, who=user, what=self)


    def wasSolvedBy(self, user):
        """Checks if this exercise has previously been solved by the user.

        """
        thisExercise = _Solution.what == self
        byThisUser = _Solution.who == user
        condition = q.AND(thisExercise, byThisUser)
        return self.store.query(_Solution, condition, limit=1).count() == 1



class _Solution(item.Item):
    """A log of an exercise being solved.

    """
    who = attributes.reference(allowNone=False)
    what = attributes.reference(allowNone=False)



def solveAndNotify(proto, exercise):
    """The user at the given AMP protocol has solved the given exercise.

    This will log the solution and notify the user.
    """
    exercise.solvedBy(proto.user)
    proto.callRemote(ce.NotifySolved,
                     identifier=exercise.identifier,
                     title=exercise.title)



class Locator(amp.CommandLocator):
    @ce.GetExercises.responder
    def getExercises(self, solved):
        return {"exercises": [{b"title": e.title, b"identifier": e.identifier}
                              for e in self._getExercises(solved)]}


    def _getExercises(self, yieldSolved):
        for ex in self.store.query(Exercise):
            if ex.wasSolvedBy(self.user):
                if yieldSolved:
                    yield ex
            else:
                if not yieldSolved:
                    yield ex



    @ce.GetExerciseDetails.responder
    def getExerciseDetails(self, identifier):
        """Gets the details for a particular exercise.

        """
        exercise = self._getExercise(identifier)
        response = {
            b"identifier": exercise.identifier,
            b"title": exercise.title,
            b"description": exercise.description,
            b"solved": exercise.wasSolvedBy(self.user)
        }
        return response


    def _getExercise(self, identifier):
        try:
            cond = Exercise.identifier == identifier
            return self.store.findUnique(Exercise, cond)
        except ItemNotFound:
            raise ce.UnknownExercise()



class SolvableResourceMixin(object):
    """A mixin for resources that know when an exercise is solved.

    The resource must be served over a multiplexed connection.
    """
    exerciseIdentifier = None

    def __init__(self, store):
        self.store = store


    def solveAndNotify(self, request):
        """Notifies the owner of the current request (so, the user doing the
        exercise) that they've solved the exercise, and mark it as
        solved in the database.

        """
        remote = request.transport.remote
        withThisIdentifier = Exercise.identifier == self.exerciseIdentifier
        exercise = self.store.findUnique(Exercise, withThisIdentifier)
        solveAndNotify(remote, exercise)



class Secret(item.Item):
    """Some secret stored entropy for a user.

    Used to randomize exercises specific to the user. This exists for
    two reasons:

    - Allowing randomizations for a particular user to be consistent
    - Preventing users from requesting arbitrary amounts of random
      data

    """
    user = attributes.reference(allowNone=False, indexed=True)
    entropy = attributes.bytes(defaultFactory=lambda: os.urandom(256 // 8))

    @classmethod
    def forUser(cls, user):
        """Finds or creates a Secret for this user.

        """
        return user.store.findOrCreate(cls, user=user)



from axiom import attributes, item
from axiom.errors import ItemNotFound
from clarent.certificate import SecureCiphersContextFactory
from OpenSSL.SSL import Context, VERIFY_PEER, SSLv23_METHOD
from OpenSSL.SSL import OP_SINGLE_DH_USE, OP_NO_SSLv2, OP_NO_SSLv3
from twisted.python import log


class User(item.Item):
    """A user.

    """
    email = attributes.bytes(allowNone=False, indexed=True)
    digest = attributes.bytes()



class UserMixin(object):
    """A mixin that makes the user available based on the peer certificate.

    Mostly intended for command locators (that are also protocols).
    Expects a ``store`` attribute, which is the store that has all of
    the user data, as well as a ``transport`` attribute, which is the
    transport that has the peer certificate we're authenticating with.

    """
    _user = None

    @property
    def user(self):
        """The current user.

        This property is cached in the ``_user`` attribute.
        """
        if self._user is not None:
            return self._user

        cert = self.transport.getPeerCertificate()
        self._user = user = userForCert(self.store, cert)
        return user



class _TOFUContextFactory(object):
    def __init__(self, store):
        self.store = store


    def getContext(self):
        """Creates a context.

        This will make contexts using ``SSLv23_METHOD``. This is
        because OpenSSL thought it would be a good idea to have
        ``TLSv1_METHOD`` mean "only use TLSv1.0" -- specifically, it
        disables TLSv1.2. Since we don't want to use SSLv2 and v3, we
        set OP_NO_SSLv2|OP_NO_SSLv3. Additionally, we set
        OP_SINGLE_DH_USE.

        """
        ctx = Context(SSLv23_METHOD)
        ctx.use_certificate_file("cert.pem")
        ctx.use_privatekey_file("key.pem")
        ctx.load_tmp_dh("dhparam.pem")
        ctx.set_options(OP_SINGLE_DH_USE|OP_NO_SSLv2|OP_NO_SSLv3)
        ctx.set_verify(VERIFY_PEER, self._verify)
        return ctx


    def _verify(self, connection, cert, errorNumber, errorDepth, returnCode):
        """Verify a certificate.

        """
        try:
            user = userForCert(self.store, cert)
        except ItemNotFound:
            log.msg("Connection attempt by {0!r}, but no user with that "
                    "e-mail address was found, cert digest was {1}"
                    .format(emailForCert(cert), cert.digest("sha512")))
            return False

        digest = cert.digest("sha512")
        if user.digest is None:
            user.digest = digest
            log.msg("First connection by {0!r}, stored digest: {1}"
                    .format(user.email, digest))
            return True
        elif user.digest == digest:
            log.msg("Successful connection by {0!r}".format(user.email))
            return True
        else:
            log.msg("Failed connection by {0!r}; cert digest was {1}, "
                    "expecting {2}".format(user.email, digest, user.digest))
            return False



class ContextFactory(object):
    def __init__(self, store):
        self._wrapped = _TOFUContextFactory(store)
        self._wrapper = SecureCiphersContextFactory(self._wrapped)


    def getContext(self):
        return self._wrapper.getContext()



def emailForCert(cert):
    """Extracts the e-mail address from the X509 data in the given cert.

    """
    return cert.get_subject().emailAddress.encode("utf-8")


def userForCert(store, cert):
    """Gets the user for the given certificate.

    """
    return store.findUnique(User, User.email == emailForCert(cert))
